/********************************** (C) COPYRIGHT *****************************
 * File Name          : main.c
 * Author             : WCH
 * Version            : V1.1
 * Date               : 2024/09/02
 * Description        : Main program body.
 ******************************************************************************
 * Copyright (c) 2021 Nanjing Qinheng Microelectronics Co., Ltd.
 * Attention: This software (modified or not) and binary are used for
 * microcontroller manufactured by Nanjing Qinheng Microelectronics.
 ******************************************************************************/

#include "stm32f10x.h"
#include "debug.h"
#include "CH390.h"
#include "CH390_Interface.h"

#include "netif/ethernetif.h"
#include "lwip/timeouts.h"
#include "lwip/dhcp.h"
#include "lwip/prot/dhcp.h"

#include "tcp_echo.h"
#include "udp_echo.h"

#define USE_DHCP 1

/**
 * @name ch390_int_handler
 * @brief Handle CH390 interrupt events, include packet receive
 */
void ch390_int_handler()
{
    uint8_t int_status = ch390_get_int_status();
    // Link status changed
    if (int_status & ISR_LNKCHG)
    {
        Delay_Ms(65);
        if (ch390_get_link_status())
        {
            printf("netif link up\r\n");
            netif_set_link_up(&ch390_netif);
        }
        else
        {
            printf("netif link down\r\n");
            netif_set_link_down(&ch390_netif);
        }
        ch390_write_reg(CH390_ISR, ISR_LNKCHG);
    }
    // Receive overflow
    if (int_status & ISR_ROS)
    {
        printf("Receive overflow\r\n");
        struct ethernetif *ethernetif = ch390_netif.state;
        do {
            ethernetif_input(&ch390_netif);
        } while (ethernetif->rx_len != 0);
    }
    // Receive overflow counter overflow
    if (int_status & ISR_ROO) printf("Overflow counter overflow\r\n");
    // Packet transmitted
    //    if(int_status & ISR_PT) printf("Packet sent\r\n");
    // Packet received
    if (int_status & ISR_PR)
    {
        struct ethernetif *ethernetif = ch390_netif.state;
        do {
            ethernetif_input(&ch390_netif);
        } while (ethernetif->rx_len != 0);
    }
}

/**
 * @name ch390_print_info
 * @brief Read and print chip's ID and MAC address
 */
void ch390_print_info()
{
    uint8_t i;
    uint8_t mac_addr[6];
    uint16_t vid = ch390_get_vendor_id();
    uint16_t pid = ch390_get_product_id();
    printf("ID: %04x%04x\r\n", vid, pid);

    ch390_get_mac(mac_addr);
    printf("MAC: ");
    for (i = 0; i < 6; i++)
    {
        printf("%02X ", mac_addr[i]);
    }
    printf("\r\n");
}

void rtc_init()
{
    uint8_t temp = 0;
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_PWR | RCC_APB1Periph_BKP, ENABLE);
    PWR_BackupAccessCmd(ENABLE);

    BKP_DeInit();
    RCC_LSEConfig(RCC_LSE_ON);
    while (RCC_GetFlagStatus(RCC_FLAG_LSERDY) == RESET && temp < 250)
    {
        temp++;
        Delay_Ms(20);
    }
    RCC_RTCCLKConfig(RCC_RTCCLKSource_LSE);
    RCC_RTCCLKCmd(ENABLE);
    RTC_WaitForLastTask();
    RTC_WaitForSynchro();
    RTC_WaitForLastTask();
    RTC_EnterConfigMode();
    RTC_SetPrescaler(32);
    RTC_WaitForLastTask();
    RTC_SetCounter(0);
    RTC_ExitConfigMode();
}

/**
 * @brief  Main program
 * @param  None
 * @retval None
 */
int main(void)
{
    struct ip4_addr ipaddr, netmask, gateway;

    Delay_Init();
    USART_Printf_Init(115200);

    rtc_init();

    ch390_gpio_init();
#ifdef CH390_INTERFACE_SPI
    ch390_spi_init();
#endif
    ch390_hardware_reset();
    // Don't read/write CH390 registers within 10ms after power on
    Delay_Ms(10);

    IP4_ADDR(&ipaddr, 192, 168, 1, 120);
    IP4_ADDR(&netmask, 255, 255, 255, 0);
    IP4_ADDR(&gateway, 192, 168, 1, 1);

    init_lwip_netif(&ipaddr, &netmask, &gateway);
    ch390_print_info();
    // Wait until phy link up
    while (!ch390_get_link_status())
    {
        Delay_Ms(100);
    }
    netif_set_link_up(&ch390_netif);
    printf("netif link up\r\n");

#if USE_DHCP
    dhcp_start(&ch390_netif);
    struct dhcp *dhcp;
    dhcp = netif_dhcp_data(&ch390_netif);
    // Wait untill dhcp complete
    while (dhcp->state != DHCP_STATE_BOUND)
    {
        if (ch390_get_int_pin())
        {
            ch390_int_handler();
        }
        sys_check_timeouts();
    }
    printf("DHCP complete\r\n");
    printf("IP: %s\r\n", ip4addr_ntoa(netif_ip4_addr(&ch390_netif)));
#endif

    udpecho_init();
    tcp_client_init();
    tcp_server_init();

    while (1)
    {
        if (ch390_get_int_pin())
        {
            ch390_int_handler();
        }

        sys_check_timeouts();
    }
}
