/********************************** (C) COPYRIGHT *******************************
 * File Name          : main.c
 * Author             : WCH
 * Version            : V1.1
 * Date               : 2024/08/20
 * Description        : Main program body.
 *********************************************************************************
 * Copyright (c) 2021 Nanjing Qinheng Microelectronics Co., Ltd.
 * Attention: This software (modified or not) and binary are used for
 * microcontroller manufactured by Nanjing Qinheng Microelectronics.
 *******************************************************************************/

#include "stm32f10x.h"
#include "debug.h"
#include "CH390.h"
#include "CH390_Interface.h"

// phy link status
uint8_t phy_linked = 0;

// Random packet data for send test
#define TEST_DATA_LEN 600
uint8_t packet_data[TEST_DATA_LEN];
uint8_t receive_buff[CH390_PKT_MAX];

void init_packet_data()
{
    int i;
    for(i = 0; i < TEST_DATA_LEN; i++)
    {
        packet_data[i] = i;
    }
}

void print_packet(uint8_t *buff, uint16_t length)
{
		int i;
    printf("------------------\r\n"
           "Packet length: %d", length);

    for(i = 0; i < length; i++)
    {
        if(i % 16 == 0)
        {
            printf("\r\n%04x   ", i);
        }
        printf("%02x ", buff[i]);
    }

    printf("\r\n");
}

/**
 * @name ch390_int_handler
 * @brief Handle CH390 interrupt events, include packet receive
 */
void ch390_int_handler()
{
    uint8_t int_status = ch390_get_int_status();
    // Link status changed
    if(int_status & ISR_LNKCHG)
    {
        Delay_Ms(65);
        phy_linked = ch390_get_link_status();
        printf("Link status: %d\r\n", phy_linked);
        ch390_write_reg(CH390_ISR, ISR_LNKCHG);
    }
    // Receive overflow
    if(int_status & ISR_ROS) 
    {
        printf("Receive overflow\r\n");
        uint8_t rx_status = 0;
        uint32_t length = 0;
        while((length = ch390_receive_packet(receive_buff, &rx_status)) != 0)
        {
            // printf("Rx: %d\r\n", length);
            print_packet(receive_buff, length);
        }
    }
    // Receive overflow counter overflow
    if(int_status & ISR_ROO) printf("Overflow counter overflow\r\n");
    // Packet transmitted
//    if(int_status & ISR_PT) printf("Packet sent\r\n");
    // Packet received
    if(int_status & ISR_PR)
    {
        /* Multiple packets may be received in a single packet receive
         * event. So "ch390_receive_packet" should be called until the
         * return value is 0. Otherwise data will accumulate in RX SRAM
         * and cause overflow. */
        uint8_t rx_status = 0;
        uint32_t length = 0;
        while((length = ch390_receive_packet(receive_buff, &rx_status)) != 0)
        {
            /* Notice, the last 4 bytes of the received data are CRC */
            // printf("Rx: %d\r\n", length);
            print_packet(receive_buff, length);
        }
    }
}

/**
 * @name ch390_print_info
 * @brief Read and print chip's ID and MAC address
 */
void ch390_print_info()
{
    uint8_t i;
    uint8_t mac_addr[6];
    uint16_t vid = ch390_get_vendor_id();
    uint16_t pid = ch390_get_product_id();
    printf("ID: %04x%04x\r\n", vid, pid);

    ch390_get_mac(mac_addr);
    printf("MAC: ");
    for (i = 0; i < 6; i++)
    {
        printf("%02X ", mac_addr[i]);
    }
    printf("\r\n");
}

/**
 * @brief  Main program
 * @param  None
 * @retval None
 */
int main(void)
{
    Delay_Init();
    USART_Printf_Init(115200);
    ch390_gpio_init();
#ifdef CH390_INTERFACE_SPI
    ch390_spi_init();
#endif
    ch390_hardware_reset();
    // Don't read/write CH390 registers within 10ms after power on
    Delay_Ms(10);
    ch390_default_config();
    ch390_print_info();

    init_packet_data();
    while (1)
    {
        if (ch390_get_int_pin())
        {
            ch390_int_handler();
        }

        if (phy_linked)
        {
            ch390_send_packet(packet_data, TEST_DATA_LEN);
            Delay_Ms(5);
        }
    }
}
