/********************************** (C) COPYRIGHT *******************************
 * File Name          : main.c
 * Author             : WCH
 * Version            : V1.1
 * Date               : 2024/08/20
 * Description        : Main program body.
 *********************************************************************************
 * Copyright (c) 2021 Nanjing Qinheng Microelectronics Co., Ltd.
 * Attention: This software (modified or not) and binary are used for
 * microcontroller manufactured by Nanjing Qinheng Microelectronics.
 *******************************************************************************/

#include "stm32f10x.h"
#include "debug.h"
#include "CH390.h"
#include "CH390_Interface.h"

#ifdef CH390_INTERFACE_16_BIT
#error WOL is not supported in 16-bit mode!
#endif

uint8_t receive_buff[CH390_PKT_MAX];
uint32_t receive_count = 0;

void print_packet(uint8_t *buff, uint16_t length)
{
		int i;
    printf("------------------\r\n"
           "Packet length: %d", length);

    for(i = 0; i < length; i++)
    {
        if(i % 16 == 0)
        {
            printf("\r\n%04x   ", i);
        }
        printf("%02x ", buff[i]);
    }

    printf("\r\n");
}

/**
 * @name ch390_int_handler
 * @brief Handle CH390 interrupt events, include packet receive
 */
void ch390_int_handler()
{
    uint8_t int_status = ch390_get_int_status();
    // Link status changed
    if(int_status & ISR_LNKCHG)
    {
        Delay_Ms(65);
        printf("Link status: %d\r\n", ch390_get_link_status());
        ch390_write_reg(CH390_ISR, ISR_LNKCHG);
    }
    // Receive overflow
    if(int_status & ISR_ROS) printf("Receive overflow\r\n");
    // Receive overflow counter overflow
    if(int_status & ISR_ROO) printf("Overflow counter overflow\r\n");
    // Packet transmitted
//    if(int_status & ISR_PT) printf("Packet sent\r\n");
    // Packet received
    if(int_status & ISR_PR)
    {
        /* Multiple packets may be received in a single packet receive
         * event. So "ch390_receive_packet" should be called until the
         * return value is 0. Otherwise data will accumulate in RX SRAM
         * and cause overflow. */
        uint8_t rx_status = 0;
        uint32_t length = 0;
        while((length = ch390_receive_packet(receive_buff, &rx_status)) != 0)
        {
            /* Notice, the last 4 bytes of the received data are CRC */
            // printf("Rx: %d\r\n", length);
            print_packet(receive_buff, length);
            receive_count++;
        }
    }
}

/**
 * @name ch390_print_info
 * @brief Read and print chip's ID and MAC address
 */
void ch390_print_info()
{
    uint8_t i;
    uint8_t mac_addr[6];
    uint16_t vid = ch390_get_vendor_id();
    uint16_t pid = ch390_get_product_id();
    printf("ID: %04x%04x\r\n", vid, pid);

    ch390_get_mac(mac_addr);
    printf("MAC: ");
    for (i = 0; i < 6; i++)
    {
        printf("%02X ", mac_addr[i]);
    }
    printf("\r\n");
}

/**
 * @name sample_frame_filter_config
 * @brief Configure wake up frame filter
 *
 * Sample frame example:
 * xx xx xx xx xx xx  xx xx xx xx xx xx  // 12 bytes MAC
 * 08 00 00 01 02 03 04 05  06 07 08 09 0a 0b 0c 0d
 * 0e 0f 10 11 12 13 14 15  16 17 18 19 1a 1b 1c 1d
 * ...
 *
 * filter0: Byte mask = 0x000000FF;  Command = 0x1;  Offset = 12;  CRC16 = 0x1D0F
 *      ... 08 00 00 01 02 03 04 05  06 07 08 09 0a 0b 0c 0d  0e 0f 10 11 12 13 14 15  16 17 18 19 1a 1b 1c 1d
 *         |-----------------------
 *
 * filter1: Byte mask = 0x0000FFFF;  Command = 0x1;  Offset = 13;  CRC16 = 0xCA3A
 *      ... 08 00 00 01 02 03 04 05  06 07 08 09 0a 0b 0c 0d  0e 0f 10 11 12 13 14 15  16 17 18 19 1a 1b 1c 1d
 *            |-------------------------------------------------
 *
 * filter2: Byte mask = 0x00FF0000;  Command = 0x1;  Offset = 14;  CRC16 = 0xE6DC
 *      ... 08 00 00 01 02 03 04 05  06 07 08 09 0a 0b 0c 0d  0e 0f 10 11 12 13 14 15  16 17 18 19 1a 1b 1c 1d
 *               |                                                  ------------------------
 *
 * filter3: Byte mask = 0x0FF0FF00;  Command = 0x1;  Offset = 15;  CRC16 = 0xE8CB
 *      ... 08 00 00 01 02 03 04 05  06 07 08 09 0a 0b 0c 0d  0e 0f 10 11 12 13 14 15  16 17 18 19 1a 1b 1c 1d
 *                  |                         ------------------------             ------------------------
 */
void sample_frame_filter_config()
{
    // Mask
    ch390_write_eeprom(0, 0x00ff);
    ch390_write_eeprom(1, 0x0000);
    ch390_write_eeprom(2, 0xffff);
    ch390_write_eeprom(3, 0x0000);
    ch390_write_eeprom(4, 0x0000);
    ch390_write_eeprom(5, 0x00ff);
    ch390_write_eeprom(6, 0xff00);
    ch390_write_eeprom(7, 0x0ff0);

    // Command
    ch390_write_eeprom(8, 0x0101);
    ch390_write_eeprom(9, 0x0101);

    // Offset
    ch390_write_eeprom(10, 0x0d0c);
    ch390_write_eeprom(11, 0x0f0e);

    // CRC16
    ch390_write_eeprom(12, 0x1d0f);
    ch390_write_eeprom(13, 0xca3a);
    ch390_write_eeprom(14, 0xe6dc);
    ch390_write_eeprom(15, 0xe8cb);
}

/**
 * @brief  Main program
 * @param  None
 * @retval None
 */
int main(void)
{
    Delay_Init();
    USART_Printf_Init(115200);
    ch390_gpio_init();
    ch390_interrupt_init();
#ifdef CH390_INTERFACE_SPI
    ch390_spi_init();
#endif
    ch390_hardware_reset();
    // Don't read/write CH390 registers within 10ms after power on
    Delay_Ms(10);
    ch390_default_config();
    ch390_print_info();

    sample_frame_filter_config();
    // Enable magic packet and sample packet wake up
    ch390_wakeup_config(WCR_MAGICEN | WCR_SAMPLEEN);

    while (1)
    {
        if (ch390_get_int_pin())
        {
            ch390_int_handler();
        }

        // Sleep every 5 packets received
        if(receive_count >= 5)
        {
            receive_count = 0;
            ch390_wake_notify();
            printf("Sleep\r\n");
            __WFI();
            printf("Wake on LAN!\r\n");
        }
    }
}

/**
 * @brief This function handles WOL Event.
 */
void EXTI9_5_IRQHandler(void)
{
    if (EXTI_GetITStatus(EXTI_Line8) != RESET)
    {
        ch390_read_reg(CH390_NSR); // Clear wake up flag
        EXTI_ClearITPendingBit(EXTI_Line8);
    }
}
